# python code for:
# Jonathan Barth, Oliver Richters: Demand driven ecological collapse: A
# stock-flow fund-service model of money, energy and ecological scale.
# In: Samuel Decker, Wolfram Elsner & Svenja Flechtner (eds.): Principles and
# Pluralist Approaches in Teaching Economics: Towards a Transformative Science.
# Routledge, July 2019, ISBN 978-1-1380-3768-7.

from pylab import exp, log, arctan, pi, plot, legend, savefig, arange, plt, zeros, array, linspace, argmin
from numpy import linalg as LA
import matplotlib

matplotlib.rcParams.update({'font.size': 13})

### 1. plots of time evolution

steps = 400

theta = 0.8 # tax rate
p = 100. # price
r = 0.05 # interest rate
a = 0.1 # ecol growth rate
cv = 0.05 # consumption out of wealth
cy = 0.8 # consumption out of income
gamma = 1.1 # transfer factor

# definition of the model
def model(S_max = 100., S_init = 100, M_init = 10., Y_init=1., G_init=4., theta= 0.5, p = 4., r = 0.1, a = 0.1, cv = 0.05, cy= 0.8, gamma = 1.1, s_target = 0.7):
    #initialization
    s=zeros(steps)
    s[0] = S_init/S_max
    M=zeros(steps)
    M[0] = M_init
    Y = zeros(steps)
    Y[0] = Y_init
    G = G_init

    # loop
    for t in arange(1,steps):
        if t==2: print(s[t-1], M[t-1], p*Y[t-1])
        Y[t] = 1./(    p/( cy*(1-theta)*(p*Y[t-1] + G) + (cv + cy*(1-theta)*r)*M[t-1]      )    +   gamma/s[t-1]/S_max    )
        s[t] = s[t-1]*(1 + a - a*s[t-1]) - gamma * Y[t] / S_max
        M[t] = (1 + (1-theta)*r) *M[t-1] - theta*p*Y[t] + (1-theta)*G

    return M, s*S_max, p*Y

# three model runs
M_stable, S_stable, Y_money_stable = model(cv = 0.06)

M_unstable, S_unstable, Y_money_unstable = model(cv = 0.04)

M_explode, S_explode, Y_money_explode = model(cv = 0.01)


# three graphs in one figure
fig1, (ax1, ax2, ax3) = plt.subplots(nrows=1, ncols=3, sharey=True, figsize=(10,4.6))

ax1.plot(arange(steps), M_stable, 'k--', label = r"stock of bills $M$")
ax1.plot(arange(steps), S_stable, 'g-.', label=r"ecosystem $S$")
ax1.plot(arange(steps), Y_money_stable, 'r-', label=r"production $Y^p p$")

ax2.plot(arange(steps), M_unstable, 'k--', label = r"stock of bills $M$")
ax2.plot(arange(steps), S_unstable, 'g-.', label=r"ecosystem $S$")
ax2.plot(arange(steps), Y_money_unstable, 'r-', label=r"production $Y^p p$")
ax2.set_xlabel(r"$t$")

ax3.plot(arange(steps), M_explode, 'k--', label = r"stock of bills $M$")
ax3.plot(arange(steps), S_explode, 'g-.', label=r"ecosystem $S$")
ax3.plot(arange(steps), Y_money_explode, 'r-', label=r"production $Y^p p$")

# cosmetics
for i in (ax1, ax2, ax3):
    i.axhline(ls=':',xmin=0,xmax=1,color='0.2',y=0, linewidth=0.5)
    i.axhline(ls=':',xmin=0,xmax=1,color='0.2',y=100, linewidth=0.5)
    
# save pdf

ax3.set_ylim(ymin=-10,ymax=110)
ax1.legend(loc="center left")
ax1.set_title("stable")
ax2.set_title("mon. stable, ecol. unstable")
ax3.set_title("explosive")
fig1.savefig("paper-time_evolution_three.pdf")
fig1.savefig("paper-time_evolution_three.png")




### 2. plot of logistic growth and constraints


fig2, (ax21, ax22) = plt.subplots(nrows=1, ncols=2, sharey=False, figsize=(10,4.6))

def logistic_time_evolution(start = 0.1, growth_rate = 0.1, maximum=100, steps=100):
    time = zeros(logistic_growth_steps)
    time[0]=start
    for t in range(1,steps):
        time[t]=time[t-1] * (1+growth_rate * (1 - time[t-1]/maximum))
    return time

logistic_growth_steps = 100
ax21.plot(range(logistic_growth_steps), logistic_time_evolution(start=10, steps=logistic_growth_steps), 'k--', label = r"start: 10")
ax21.plot(range(logistic_growth_steps), logistic_time_evolution(start=0.2, steps=logistic_growth_steps), 'k-', label = r"start: 0.2")
ax21.set_xlabel(r"time $t$")
ax21.set_ylabel(r"biomass $S$ with $S_{max} = 100$")
ax21.legend(loc="lower right")
ax21.axhline(ls=':',xmin=0,xmax=1,color='0.2',y=100, linewidth=0.5)
ax21.axhline(ls=':',xmin=0,xmax=1,color='0.2',y=0, linewidth=0.5)
ax21.set_title("logistic growth")

Y_target_array = arange(0,5,0.01)
Y_realized_array = Y_target_array/(1+Y_target_array)

ax22.set_title("ecological constraint")
ax22.plot(Y_target_array , Y_realized_array, 'k-', label=r"realized harvest $E$")
ax22.plot(Y_target_array , Y_target_array, 'k--', label=r"targeted harvest $E^\top$")
ax22.set_xlabel(r"targeted harvest $E^\top$")
ax22.set_ylabel(r"harvest")
ax22.axis('equal')
ax22.axhline(ls='-.',xmin=0,xmax=1,color='0.2',y=1, linewidth=1, label="maximum harvest")
ax22.axhline(ls=':',xmin=0,xmax=1,color='0.2',y=0, linewidth=0.5)
ax22.legend(loc="upper left")


fig2.savefig("paper-logistic_growth_and_constraints.pdf")
fig2.savefig("paper-logistic_growth_and_constraints.png")


### 3. plot of bifurcation diagram


fig3 = plt.figure()
ax3 = fig3.add_subplot(111)

def bifurcation_diagramm(S_max = 100., gammae = 1.1, a=0.1, cy=0.8, theta=0.5, p=4., G=4.):
    S_array=linspace(0,S_max,1001)
    cr1 = array([ - a*S*(S_max-S)*(S*a*cy - S_max*a*cy + S_max*cy - S_max)*(1-theta)*p  /  (-a*S**2*theta*p + S*S_max*a*p*theta + G*S_max*gammae*theta-G*S_max*gammae) / (a*S-S_max*a+S_max)   for S in S_array])
    cr2 = [ a*S*(S_max-S)*(S_max/(S_max-a*(S_max-S)) - cy ) / (theta/(1-theta)*(S_max-S)*S*a - G * gammae * S_max/p)  for S in S_array]
    return S_array, cr1, cr2

S, cr1, cr2 = bifurcation_diagramm()

S_clean =  S[cr1>0]
cr1_clean =  cr1[cr1>0]

min_cr1_index = argmin(cr1_clean)
cr1_min = cr1_clean[min_cr1_index]
S_at_cr1_min = S_clean[min_cr1_index]

cr1_stabiler_ast = cr1_clean[S_clean>S_at_cr1_min]
S_stabiler_ast = S_clean[S_clean>S_at_cr1_min]
cr1_instabiler_ast = cr1_clean[S_clean<S_at_cr1_min]
S_instabiler_ast = S_clean[S_clean<S_at_cr1_min]

ax3.plot(cr1_stabiler_ast, S_stabiler_ast,'-', label="stable fixed points")
ax3.plot(cr1_instabiler_ast, S_instabiler_ast,'-', label="unstable fixed points")
ax3.plot(cr1_min, S_at_cr1_min ,'ok', label="limit point")
ax3.axvline(ls=":",ymin=0,ymax=1,color='0.2',x=cr1_min, label="ecological stability frontier")
ax3.legend(loc="center right")
ax3.set_xlim(xmin=0,xmax=2.5)
ax3.set_ylim(ymin=0,ymax=100)


ax3.set_xlabel(r"$c_r = c_m / r$")
ax3.set_ylabel(r"$S$ with $S_{max} = 100$")

fig3.savefig("paper-bifurcation-diagramm.pdf")
fig3.savefig("paper-bifurcation-diagramm.png")


### 4. plot stability frontiers


fig4, ax4 = plt.subplots()

r_array=linspace(0,0.4,101)
cv_array=linspace(0,0.2,101)
bifurcation_lines = [[1000]+[cv/r for r in r_array[1:]] for cv in cv_array]

ax4.set_xticks(linspace( r_array[0], r_array[-1],5), minor=False) #+diffcalc/2.
ax4.set_yticks(linspace(cv_array[0],cv_array[-1],5), minor=False) #+diffcalc/2.

ax4.text(r_array[-1]*0.98, cv_array[-1]*0.97, r'stable fixed point', color = 'k', ha="right", va="top", bbox = dict(ec='1',fc='1'), rotation=31)
ax4.text(r_array[-1]*0.96, cv_array[-1]*0.56, r'monetarily stable,',  color = 'k', ha="right", va="center", bbox = dict(ec='1',fc='1'), rotation=23)
ax4.text(r_array[-1]*0.98, cv_array[-1]*0.46, r'ecologically unstable',  color = 'k', ha="right", va="center", bbox = dict(ec='1',fc='1'), rotation=23)
ax4.text(r_array[-1]*0.98, cv_array[-1]/10, r'monetarily and ecologically unstable', color = 'k', ha="right", va="center", bbox = dict(ec='1',fc='1'), rotation=5)

ax4.text(r_array[40], 0.44*r_array[40], r"$c_r = c_m/r = 0.44$", color="blue", ha="center", va="center", bbox = dict(ec='1',fc='1'), rotation=30)
ax4.text(r_array[60], 0.2*r_array[60], r"$c_r = 0.20$", color="red", ha="center", va="center", bbox = dict(ec='1',fc='1'), rotation=16)

# steuerlinesplot = ax4.contour(r_array, cv_array, bifurcation_lines, [0.2, 0.44], linestyles= ['dashed', 'solid'], colors=['red','blue'], label=["ecological","monetary"])
# clabel(steuerlinesplot, fontsize=15, inline=1, inline_spacing=25, fmt=r"$c_r = c_m/r = %1.2f$", manual=[(0.25,0.05),(0.17,0.08)]) 

ax4.plot(r_array, r_array*0.44, 'b-', label="ecological stability frontier")
ax4.plot(r_array, r_array*0.2, 'r--', label="monetary stability frontier")
ax4.set_xlim(xmin=r_array[0], xmax=r_array[-1])
ax4.set_ylim(ymin=cv_array[0], ymax=cv_array[-1])

ax4.set_xlabel(r"interest rate $r$")
ax4.set_ylabel(r"consumption out of wealth $c_m$")


ax4.legend(loc="upper left")
fig4.savefig("paper-stability-frontiers.pdf")
fig4.savefig("paper-stability-frontiers.png")
